// h8300fix.idc
// This script seeks out unresolved issues and attempts to create fixups.  It is primarily for
//   use with imported h8 elf files, although it can be used with images as well.
//   1) "jsr @r?"'s and attempts to create valid code xrefs
//   2) "jsr @@"'s vectors are forced to words-offsets
//
// History:
// __before__  twh
//             - initial coding and testing
//             - removed a lot of stuff, due to jump tables now being supported by the h8 processor module
//             saf/twh
//             - enhanced LookBack() to support "floating targets"
// 03/12/2001  twh
//             - first public release
// 03/15/2001  twh
//             - oops, now arrays are deleted when not needed, to free up memory

#include <idc.idc>

// this may vary between 0x0000 and 0x10000 (elf file vs. hex image, respectively)
#define SEGOFFSET   0x0000

// locations where you expect absolute (double) indirects to appear
#define ABS8_START  0x0030
#define ABS8_END    0x00FE

#define TRUE        1
#define FALSE       0

// the only time you might need this is if the compiler generates unused ABS$8 entries...
//    ours still does... sigh
// IDA now handles xrefs for these, if they are referenced somewhere in the image
static FixupDblInd(void)
{
    auto ea;
    auto vea;

    for (ea = (SEGOFFSET + ABS8_START); ea <= (SEGOFFSET + ABS8_END); ea = ea + 2 )
	{
		if ( strstr(Name(ea), "_$ind_opt") == 0 )
		{
			MakeWord(ea);
			OpOff(ea, 0, SEGOFFSET);
		}
    }

}


static CreateNewList( szArrayName )
{
    auto MyArray;

    // create an array to hold long vals
    MyArray =  CreateArray( szArrayName );
    if (MyArray == -1)
    {
        MyArray = GetArrayId( szArrayName );
        DeleteArray( MyArray );
        MyArray =  CreateArray( szArrayName );
    }
    return MyArray;
}


static AddToList( MyArray, ea)
{
    auto i, j;

    j = GetLastIndex(AR_LONG, MyArray);

    // check if the current location is already stored in the array
    for (i = 0; i <= j; i++ )
    {
        if (ea == GetArrayElement(AR_LONG, MyArray, i))
        {
            // location is present
            return FALSE;
        }
    }

    // location is not present; add it to the array
    SetArrayLong( MyArray, j+1, ea );
    return TRUE;
}


// a very messy recursive backward code flow search function, sorry
static LookBack(ea, lbaid, oea, flag)
{
    auto xea;
    auto zea;
	auto flType;
    auto lbaid1;

    // Check all code xrefs
    if (ea != BADADDR)
    {
        xea = RfirstB(ea);
        while (xea != BADADDR)
    	{
		    // We only search in the current function
			flType = XrefType();
		    if ((flType == fl_JF) || (flType == fl_JN) || (flType == fl_F))
		    {
                // Skip a location we have already visited
                if ( AddToList(lbaid, xea) )
                {
        			// Check for "mov" to wanted register number
					if (GetMnem(xea) == "mov") {
						// make sure register numbers match
						if (((GetOpnd(xea, 1) == substr(GetOpnd(oea, 0), 1, 3)) && flag) || ((GetOpnd(xea, 1) == GetOpnd(oea, 0)) && !flag))
						{
							// display progression to immediate move
							Message("        Found 0x%06x \"%s %s, %s\".\n",
								xea, GetMnem(xea), GetOpnd(xea, 0), GetOpnd(xea, 1));

        					if ( GetOpType(xea, 0) == 5 )
							{
        						return xea;
							}
							else
							{
								lbaid1 = CreateNewList("LBA" + ltoa(ea, 16));
								if (lbaid1 == -1)
								{
									// exit
									Warning("Cannot create an array: LBA" + ltoa(ea, 16) + "!");
									return BADADDR;
								}
								else
								{
									zea = LookBack(xea, lbaid1, xea, 0);
									if (zea != BADADDR)
									{
										return zea;
									}
                                    DeleteArray( lbaid1 );
								}
							}
						}
					}
    			    zea = LookBack(xea, lbaid, oea, flag);
                    if (zea != BADADDR)
                    {
                        return zea;
                    }
                }
		    }
	        xea = RnextB(ea, xea);
	    }
    }
    return BADADDR;
}


// search for and destroy jsr @r?'s  :)
static FixupInds(void)
{
	auto ea;
	auto vea;
	auto dea;
    auto lbaid;

    ea = FindCode(MinEA(), SEARCH_DOWN);

	while (ea != BADADDR)
	{
        // Check for "jsr"
        if (GetMnem(ea) == "jsr")
        {
            // Check for indirect
            if (GetOpType(ea, 0) == 3)    // 3 = indirect register
            {
                // Determine best guess at the indirect address value
			    Message("    Detected 0x%06x \"jsr %s\".\n", ea, GetOpnd(ea, 0) );
				lbaid = CreateNewList("LBA" + ltoa(ea, 16));
				if (lbaid == -1)
				{
					// exit
					Warning("Cannot create an array: LBA" + ltoa(ea, 16) + "!");
					return;
				}
                vea = LookBack(ea, lbaid, ea, 1);
                DeleteArray( lbaid );
                if (vea != BADADDR)
				{
					OpOff(vea, 0, SEGOFFSET);
					dea = GetOperandValue(vea, 0);
                    if (dea != -1)
                    {
                        AddCodeXref(ea, dea, fl_CN);
					    MakeComm(ea, "jsr " + NameEx(ea, dea));
                        Message("            Code XREF (fl_CN) was added.\n");
                    }
                    else
                    {
					    Message("**          NO CODE XREF ADDED!\n");
                    }
				}
				else
				{
					Message("    Found \"jsr %s\" @ %s ... could NOT determine target!\n",
						GetOpnd(ea, 0), ltoa(ea, 16) );
				}
            }
        }
        ea = FindCode(ea, SEARCH_DOWN);
    }
}


// perform all fixup operations
static main(void)
{
//    FixupDblInd();
    FixupInds();
    Message("h8300fix.idc execution finished.\n");
}


